﻿using System;
using System.Collections.Generic;
using System.Linq;
using Utils;
using ZWave.CommandClasses;
using ZWave.Security;

namespace ZWave.ZnifferApplication
{
    public class SecurityManager
    {
        private static object _locker = new object();
        private static List<byte[]> _securityKeys;
        private static List<byte[]> _tempSecurityKeys;
        private static Dictionary<int, SpanCache> _spanCacheDictionary;
        private static Dictionary<int, MpanCache> _mpanCacheDictionary;
        private static Dictionary<int, NonceS0Cache> _nonceS0CacheDictionary;
        private static SecurityS0CryptoProviderBase _securityS0 { get; set; }
        private static SecurityS2CryptoProviderBase _securityS2 { get; set; }
        public static MpanTable MpanTable { get; set; }
        public static SpanTable SpanTable { get; set; }
        public static Dictionary<InvariantPeerNodeId, SinglecastKey> ScKeys { get; set; }
        public static Dictionary<NodeGroupId, MulticastKey> McKeys { get; set; }

        public static void Initialize()
        {
            _securityKeys = new List<byte[]>();
            AddPresetKeys();
            _tempSecurityKeys = new List<byte[]>();
            _spanCacheDictionary = new Dictionary<int, SpanCache>();
            _mpanCacheDictionary = new Dictionary<int, MpanCache>();
            _nonceS0CacheDictionary = new Dictionary<int, NonceS0Cache>();
            MpanTable = new MpanTable();
            SpanTable = new SpanTable();
            ScKeys = new Dictionary<InvariantPeerNodeId, SinglecastKey>();
            McKeys = new Dictionary<NodeGroupId, MulticastKey>();
            _securityS0 = new SecurityS0CryptoProviderBase();
            _securityS2 = new SecurityS2CryptoProviderBase();
        }

        public static void Reset()
        {
            _securityKeys.Clear();
            AddPresetKeys();
            _tempSecurityKeys.Clear();
            _spanCacheDictionary.Clear();
            _mpanCacheDictionary.Clear();
            _syncIdCounter = 0;
            nonceS0Storage.Clear();
            spanStorage.Clear();
            mpanValueStorage.Clear();
        }

        private static void AddPresetKeys()
        {
            _securityKeys.AddRange(new[]
            {
                "01010101010101010101010101010101".GetBytes(),
                "C0000000000000000000000000000000".GetBytes(),
                "C1000000000000000000000000000000".GetBytes(),
                "C2000000000000000000000000000000".GetBytes(),
                "00000000000000000000000000000000".GetBytes(),
            });
        }

        private static Dictionary<int, Dictionary<OrdinalPeerNodeId, byte[]>> nonceS0Storage = new Dictionary<int, Dictionary<OrdinalPeerNodeId, byte[]>>();
        private static Dictionary<int, Dictionary<InvariantPeerNodeId, SpanRecord>> spanStorage = new Dictionary<int, Dictionary<InvariantPeerNodeId, SpanRecord>>();
        private static Dictionary<int, Dictionary<NodeGroupId, Dictionary<int, MpanValue>>> mpanValueStorage = new Dictionary<int, Dictionary<NodeGroupId, Dictionary<int, MpanValue>>>();

        public static void ParseSecurity(DataItem dataItem)
        {
            if (dataItem.Destination != null && dataItem.Destination.Length > 0 && dataItem.LastPayload != null && dataItem.LastPayload.Length > 1 && dataItem.HomeId != null && dataItem.HomeId.Length == 4)
            {
                if (dataItem.LastPayload[0] == COMMAND_CLASS_SECURITY.ID)
                {
                    if (dataItem.LastPayload[1] == COMMAND_CLASS_SECURITY.SECURITY_NONCE_REPORT.ID)
                    {
                        var peerNodeId = new OrdinalPeerNodeId(dataItem.Source, dataItem.Destination[0]);
                        var homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
                        COMMAND_CLASS_SECURITY.SECURITY_NONCE_REPORT cmd = dataItem.LastPayload;
                        if (!nonceS0Storage.ContainsKey(homeId))
                        {
                            nonceS0Storage.Add(homeId, new Dictionary<OrdinalPeerNodeId, byte[]>());
                        }
                        if (!nonceS0Storage[homeId].ContainsKey(peerNodeId))
                        {
                            nonceS0Storage[homeId].Add(peerNodeId, null);
                        }
                        nonceS0Storage[homeId][peerNodeId] = cmd.nonceByte.ToArray();
                    }
                    else
                        if (dataItem.LastPayload[1] == COMMAND_CLASS_SECURITY.SECURITY_MESSAGE_ENCAPSULATION.ID ||
                            dataItem.LastPayload[1] == COMMAND_CLASS_SECURITY.SECURITY_MESSAGE_ENCAPSULATION_NONCE_GET.ID)
                    {
                        var peerNodeId = new OrdinalPeerNodeId(dataItem.Destination[0], dataItem.Source);
                        var homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
                        COMMAND_CLASS_SECURITY.SECURITY_MESSAGE_ENCAPSULATION cmd = dataItem.LastPayload;
                        byte[] nonce = null;
                        if (nonceS0Storage.ContainsKey(homeId))
                        {
                            if (nonceS0Storage[homeId].ContainsKey(peerNodeId))
                            {
                                nonce = nonceS0Storage[homeId][peerNodeId];
                            }
                        }
                        if (nonce != null)
                        {
                            dataItem.AddExtension(DataItemExtensionTypes.NonceS0, nonce);
                        }
                    }
                }
                else if (dataItem.LastPayload[0] == COMMAND_CLASS_SECURITY_2.ID)
                {
                    if (dataItem.LastPayload[1] == COMMAND_CLASS_SECURITY_2.SECURITY_2_NONCE_REPORT.ID)
                    {
                        var homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
                        COMMAND_CLASS_SECURITY_2.SECURITY_2_NONCE_REPORT cmd = dataItem.LastPayload;
                        if (cmd.properties1.sos > 0)
                        {
                            var peerNodeId = new InvariantPeerNodeId(dataItem.Source, dataItem.Destination[0]);
                            if (!spanStorage.ContainsKey(homeId))
                            {
                                spanStorage.Add(homeId, new Dictionary<InvariantPeerNodeId, SpanRecord>());
                            }
                            if (!spanStorage[homeId].ContainsKey(peerNodeId))
                            {
                                spanStorage[homeId].Add(peerNodeId, null);
                            }
                            spanStorage[homeId][peerNodeId] = new SpanRecord
                            {
                                SyncId = NextSyncId(),
                                Generation = 0,
                                ReceiverValue = cmd.receiversEntropyInput.ToArray(),
                            };
                        }
                    }
                    else if (dataItem.LastPayload[1] == COMMAND_CLASS_SECURITY_2.SECURITY_2_MESSAGE_ENCAPSULATION.ID)
                    {
                        if (HeaderStore.GetIsSinglecast(dataItem.HeaderType))
                        {
                            var homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
                            COMMAND_CLASS_SECURITY_2.SECURITY_2_MESSAGE_ENCAPSULATION cmd = dataItem.LastPayload;
                            var peerNodeId = new InvariantPeerNodeId(dataItem.Source, dataItem.Destination[0]);
                            SpanRecord span = null;
                            if (spanStorage.ContainsKey(homeId))
                            {
                                if (spanStorage[homeId].ContainsKey(peerNodeId))
                                {
                                    span = spanStorage[homeId][peerNodeId];
                                }
                            }
                            if (span != null)
                            {
                                if (span.IsValidSequenceNumber(dataItem.Source, dataItem.Destination[0], cmd.sequenceNumber))
                                {
                                    span.SetSequenceNumber(dataItem.Source, dataItem.Destination[0], cmd.sequenceNumber);
                                    span.Generation++;
                                }
                                if (cmd.properties1.extension == 1)
                                {
                                    foreach (var extData in cmd.vg1)
                                    {
                                        switch (extData.properties1.type)
                                        {
                                            case (byte)ExtensionTypes.Span:
                                                if (extData.extensionLength == 18 && extData.extension.Count == 16)
                                                {
                                                    if (span.SenderValue == null || span.SenderValue.Sum(x => x) == 0)
                                                    {
                                                        span.SenderValue = extData.extension.ToArray();
                                                    }
                                                }
                                                break;
                                            case (byte)ExtensionTypes.MpanGrp:
                                                if (extData.extensionLength == 3 && extData.extension.Count == 1)
                                                {
                                                    //
                                                }
                                                break;
                                            case (byte)ExtensionTypes.Mos:
                                                if (extData.extensionLength == 2 && extData.extension.Count == 0)
                                                {
                                                    //
                                                }
                                                break;
                                            default:
                                                break;
                                        }
                                    }
                                }
                                if (span.SenderValue != null)
                                {
                                    dataItem.AddExtension(DataItemExtensionTypes.Span, span);
                                }
                            }
                        }
                    }
                }
            }
        }

        private static int _syncIdCounter = 0;
        private static int NextSyncId()
        {
            return ++_syncIdCounter;
        }

        public static bool DecryptS0(DataItem dataItem, out byte[] decryptKey, out byte[] payload, out byte decryptedProperties)
        {
            bool isDecrypted = false;
            decryptedProperties = 0;
            payload = null;
            decryptKey = null;
            if (dataItem.Destination != null &&
                dataItem.Destination.Length > 0 &&
                dataItem.LastPayload != null &&
                dataItem.LastPayload.Length > 1 &&
                dataItem.NonceS0 != null)
            {
                try
                {
                    int homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
                    if (!_nonceS0CacheDictionary.ContainsKey(homeId))
                    {
                        _nonceS0CacheDictionary.Add(homeId, new NonceS0Cache());
                    }
                    var s0Cache = _nonceS0CacheDictionary[homeId];
                    if (_securityKeys != null)
                    {
                        var startIndex = s0Cache.NetworkKeyIndex >= 0 ? s0Cache.NetworkKeyIndex : s0Cache.ProbedNetworkKeyCount;
                        var endIndex = s0Cache.NetworkKeyIndex >= 0 ? s0Cache.NetworkKeyIndex + 1 : _securityKeys.Count;
                        for (int keyIndex = startIndex; keyIndex < endIndex; keyIndex++)
                        {
                            var key = _securityKeys[keyIndex];
                            _securityS0.OnNetworkKeyS0Changed(key);
                            _securityS0.NonceS0Storage.RegisterInternal(new OrdinalPeerNodeId(dataItem.Source, dataItem.Destination[0]), dataItem.NonceS0);
                            isDecrypted = _securityS0.DecryptFrame(dataItem.Source, dataItem.Destination[0], dataItem.HomeId, dataItem.LastPayload, dataItem.LastPayload[1], out payload, out decryptedProperties);
                            s0Cache.ProbedNetworkKeyCount = keyIndex + 1;
                            if (isDecrypted)
                            {
                                s0Cache.NetworkKeyIndex = keyIndex;
                                decryptKey = key;
                                if (key.Sum(x => x) == 0)
                                {
                                    _nonceS0CacheDictionary.Remove(homeId);
                                }
                                break;
                            }
                        }
                    }
                }
                catch
                {
                    // ignored
                }
            }
            return isDecrypted;
        }

        private static void FireNetworkKeyS2Changed(InvariantPeerNodeId peerNodeId, byte[] networkKey, ZWave.Enums.SecuritySchemes securityScheme)
        {
            var mpanKey = new byte[SecurityS2Utils.KEY_SIZE];
            var ccmKey = new byte[SecurityS2Utils.KEY_SIZE];
            var personalization = new byte[SecurityS2Utils.PERSONALIZATION_SIZE];
            if (securityScheme == ZWave.Enums.SecuritySchemes.S2_TEMP)
            {
                SecurityS2Utils.TempKeyExpand(networkKey, ccmKey, personalization, mpanKey);
            }
            else
            {
                SecurityS2Utils.NetworkKeyExpand(networkKey, ccmKey, personalization, mpanKey);
            }

            switch (peerNodeId.NodeId2)
            {
                case 0:
                    for (int ii = 0; ii <= ushort.MaxValue; ii++)
                    {
                        var index = new InvariantPeerNodeId(ii);
                        if (ScKeys.ContainsKey(index) && !ScKeys[index].CcmKey.SequenceEqual(ccmKey) &&
                            ScKeys[index].SecurityScheme == securityScheme)
                        {
                            ScKeys[index].CcmKey = ccmKey;
                            ScKeys[index].Personalization = personalization;
                            if (SpanTable.GetSpanState(index) != SpanStates.ReceiversNonce)
                            {
                                SpanTable.SetNonceFree(index);
                            }
                        }
                    }
                    for (int i = 0; i <= ushort.MaxValue; i++)
                    {
                        var index = new NodeGroupId(i);
                        if (McKeys.ContainsKey(index) && !McKeys[index].CcmKey.SequenceEqual(ccmKey) &&
                            McKeys[index].SecurityScheme == securityScheme)
                        {
                            MpanTable.RemoveRecord(index);
                        }
                    }
                    break;
                case 0xFF:
                    for (int ii = 0; ii <= ushort.MaxValue; ii++)
                    {
                        var i = new InvariantPeerNodeId(ii);
                        if (ScKeys.ContainsKey(i))
                        {
                            ScKeys[i].CcmKey = ccmKey;
                            ScKeys[i].Personalization = personalization;
                            ScKeys[i].SecurityScheme = securityScheme;
                        }
                    }
                    MpanTable.ClearMpanTable();
                    SpanTable.ClearNonceTable();
                    break;
                default:
                    if (ScKeys.ContainsKey(peerNodeId))
                    {
                        ScKeys[peerNodeId].CcmKey = ccmKey;
                        ScKeys[peerNodeId].Personalization = personalization;
                        ScKeys[peerNodeId].SecurityScheme = securityScheme;
                    }
                    else
                    {
                        ScKeys.Add(peerNodeId,
                            new SinglecastKey
                            {
                                CcmKey = ccmKey,
                                Personalization = personalization,
                                SecurityScheme = securityScheme
                            });
                    }

                    if (SpanTable.GetSpanState(peerNodeId) != SpanStates.ReceiversNonce)
                    {
                        SpanTable.SetNonceFree(peerNodeId);
                    }
                    break;
            }
        }

        private static void FireNetworkKeyS2ChangedMulti(byte groupId, byte ownerId, byte[] networkKey, ZWave.Enums.SecuritySchemes securityScheme)
        {
            var mpanKey = new byte[SecurityS2Utils.KEY_SIZE];
            var ccmKey = new byte[SecurityS2Utils.KEY_SIZE];
            var personalization = new byte[SecurityS2Utils.PERSONALIZATION_SIZE];

            SecurityS2Utils.NetworkKeyExpand(networkKey, ccmKey, personalization, mpanKey);

            var peerGroupId = new NodeGroupId(ownerId, groupId);
            if (groupId != 0)
            {
                if (McKeys.ContainsKey(peerGroupId))
                {
                    McKeys[peerGroupId].CcmKey = ccmKey;
                    McKeys[peerGroupId].MpanKey = mpanKey;
                    McKeys[peerGroupId].SecurityScheme = securityScheme;
                }
                else
                {
                    McKeys.Add(peerGroupId, new MulticastKey { CcmKey = ccmKey, MpanKey = mpanKey, SecurityScheme = securityScheme });
                }
            }
        }

        public static bool DecryptS2(DataItem dataItem, out byte[] decryptKey, out byte[] payload, out Extensions extensions)
        {
            bool isDecrypted = false;
            decryptKey = null;
            payload = null;
            extensions = null;
            var homeId = BitConverter.ToInt32(dataItem.HomeId, 0);
            COMMAND_CLASS_SECURITY_2.SECURITY_2_MESSAGE_ENCAPSULATION cmd = dataItem.LastPayload;
            try
            {
                if (cmd.vg1 != null && (HeaderStore.GetIsBroadcast(dataItem.HeaderType) || HeaderStore.GetIsMulticast(dataItem.HeaderType)))
                {
                    extensions = new Extensions();
                    extensions.ExtensionsList = cmd.vg1;
                    var mpanGrp = cmd.vg1.FirstOrDefault(x => x.properties1.type == (byte)ExtensionTypes.MpanGrp);
                    if (mpanGrp != null && mpanGrp.extension != null && mpanGrp.extension.Count == 1)
                    {
                        NodeGroupId nodeGroupId = new NodeGroupId(dataItem.Source, mpanGrp.extension[0]);
                        var mpanValue = GetMpan(dataItem.Position, dataItem.HomeId, nodeGroupId);
                        if (mpanValue != null)
                        {
                            var key = _securityKeys[mpanValue.KeyIndex];
                            FireNetworkKeyS2ChangedMulti(nodeGroupId.GroupId, nodeGroupId.NodeId, key, ZWave.Enums.SecuritySchemes.S2_ACCESS);
                            isDecrypted = DecryptS2MulticastWithCache(dataItem, nodeGroupId, mpanValue.Value, out payload, out extensions);
                            if (isDecrypted)
                            {
                            }
                        }
                    }
                }
                else if (dataItem.Destination != null && dataItem.Destination.Length > 0)
                {
                    var peerNodeId = new InvariantPeerNodeId(dataItem.Source, dataItem.Destination[0]);

                    if (dataItem.Span != null)
                    {
                        SpanRecord span = dataItem.Span;
                        if (!_spanCacheDictionary.ContainsKey(span.SyncId))
                        {
                            _spanCacheDictionary.Add(span.SyncId, new SpanCache());
                        }
                        var s2Cache = _spanCacheDictionary[span.SyncId];
                        if (s2Cache.TempNetworkKeyIndex < 0 && _securityKeys != null)
                        {
                            var startIndex = s2Cache.NetworkKeyIndex >= 0 ? s2Cache.NetworkKeyIndex : s2Cache.ProbedNetworkKeyCount;
                            var endIndex = s2Cache.NetworkKeyIndex >= 0 ? s2Cache.NetworkKeyIndex + 1 : _securityKeys.Count;
                            for (int keyIndex = startIndex; keyIndex < endIndex; keyIndex++)
                            {
                                var key = _securityKeys[keyIndex];
                                FireNetworkKeyS2Changed(peerNodeId, key, ZWave.Enums.SecuritySchemes.S2_ACCESS);
                                lock (_locker)
                                {
                                    isDecrypted = DecryptS2SinglecastWithCache(dataItem, peerNodeId, span, s2Cache, out payload, out extensions);
                                }
                                s2Cache.ProbedNetworkKeyCount = keyIndex + 1;
                                if (isDecrypted)
                                {
                                    decryptKey = key;
                                    s2Cache.NetworkKeyIndex = keyIndex;
                                    var mpanext = extensions.EncryptedExtensionsList.FirstOrDefault(x => x.properties1.type == (byte)ExtensionTypes.Mpan);
                                    if (mpanext != null && mpanext.extension != null && mpanext.extension.Count == 17)
                                    {
                                        MpanValue mpanValue = new MpanValue();
                                        mpanValue.Position = dataItem.Position;
                                        mpanValue.Value = mpanext.extension.Skip(1).ToArray();
                                        mpanValue.KeyIndex = keyIndex;
                                        RegisterMpan(dataItem.Position, dataItem.HomeId, new NodeGroupId(dataItem.Source, mpanext.extension[0]), mpanValue);
                                    }
                                    break;
                                }
                                else
                                {
                                    s2Cache.CtxSpanCacheSlices.Clear();
                                }
                            }
                        }
                        if (!isDecrypted && _tempSecurityKeys != null)
                        {
                            var startIndex = s2Cache.TempNetworkKeyIndex >= 0 ? s2Cache.TempNetworkKeyIndex : s2Cache.ProbedTempNetworkKeyCount;
                            var endIndex = s2Cache.TempNetworkKeyIndex >= 0 ? s2Cache.TempNetworkKeyIndex + 1 : _tempSecurityKeys.Count;
                            for (int keyIndex = startIndex; keyIndex < endIndex; keyIndex++)
                            {
                                var key = _tempSecurityKeys[keyIndex];
                                FireNetworkKeyS2Changed(peerNodeId, key, ZWave.Enums.SecuritySchemes.S2_TEMP);
                                lock (_locker)
                                {
                                    isDecrypted = DecryptS2SinglecastWithCache(dataItem, peerNodeId, span, s2Cache, out payload, out extensions);
                                }
                                s2Cache.ProbedTempNetworkKeyCount = keyIndex + 1;
                                if (isDecrypted)
                                {
                                    decryptKey = key;
                                    s2Cache.TempNetworkKeyIndex = keyIndex;
                                    break;
                                }
                                else
                                {
                                    s2Cache.CtxSpanCacheSlices.Clear();
                                }
                            }
                        }
                    }

                }
            }
            catch
            {
                // ignored
            }
            if (extensions == null)
            {
                extensions = new Extensions();
                extensions.ExtensionsList = cmd.vg1;
            }
            return isDecrypted;
        }

        private static MpanValue GetMpan(int position, byte[] homeId, NodeGroupId nodeGroupId)
        {
            MpanValue ret = null;
            var homeKey = BitConverter.ToInt32(homeId, 0);
            if (mpanValueStorage.ContainsKey(homeKey))
            {
                if (mpanValueStorage[homeKey].ContainsKey(nodeGroupId))
                {
                    var mpanValues = mpanValueStorage[homeKey][nodeGroupId];
                    var mpanValueKey = mpanValues.Keys;//.Keys.Where(x => x > position);
                    if (mpanValueKey.Count() > 0)
                    {
                        ret = mpanValues[mpanValueKey.First()];
                    }
                }
            }
            return ret;
        }

        internal static void RegisterMpan(int position, byte[] homeId, NodeGroupId nodeGroupId, MpanValue mpan)
        {
            var homeKey = BitConverter.ToInt32(homeId, 0);
            if (!mpanValueStorage.ContainsKey(homeKey))
            {
                mpanValueStorage.Add(homeKey, new Dictionary<NodeGroupId, Dictionary<int, MpanValue>>());
            }
            if (!mpanValueStorage[homeKey].ContainsKey(nodeGroupId))
            {
                mpanValueStorage[homeKey].Add(nodeGroupId, new Dictionary<int, MpanValue>());
            }
            if (!mpanValueStorage[homeKey][nodeGroupId].ContainsKey(position))
            {
                mpanValueStorage[homeKey][nodeGroupId].Add(position, mpan);
            }
        }

        private static bool DecryptS2MulticastWithCache(DataItem dataItem, NodeGroupId nodeGroupId, byte[] mpanValue, out byte[] payload, out Extensions extensions)
        {
            extensions = null;
            payload = null;
            var mpanContainer = MpanTable.AddOrReplace(nodeGroupId, dataItem.LastPayload[2], null, mpanValue);

            bool isDecrypted = false;
            if (McKeys.ContainsKey(nodeGroupId))
            {
                SecurityS2CryptoProviderBase.DecrementMpan(MpanTable, McKeys[nodeGroupId], nodeGroupId, new byte[16]);
                SecurityS2CryptoProviderBase.DecrementMpan(MpanTable, McKeys[nodeGroupId], nodeGroupId, new byte[16]);
                SecurityS2CryptoProviderBase.MaxMpanIterations = 1;
                for (int i = 0; i < 30; i++)
                {
                    isDecrypted = SecurityS2CryptoProviderBase.DecryptMulticastFrame(
                        MpanTable,
                        McKeys[nodeGroupId],
                        nodeGroupId,
                        dataItem.HomeId,
                        dataItem.LastPayload,
                        out payload,
                        out extensions);

                    if (isDecrypted)
                    {
                        break;
                    }
                    SecurityS2CryptoProviderBase.IncrementMpan(MpanTable, McKeys[nodeGroupId], nodeGroupId, new byte[16]);
                }
            }
            return isDecrypted;
        }

        private static bool DecryptS2SinglecastWithCache(DataItem dataItem, InvariantPeerNodeId peerNodeId, SpanRecord nonce, SpanCache spanCache, out byte[] payload, out Extensions extensions)
        {
            extensions = null;
            payload = null;
            SpanTable.Add(peerNodeId, nonce.ReceiverValue, 0, 0);
            var spanContainer = SpanTable.GetContainer(peerNodeId);
            spanContainer.InstantiateWithSenderNonce(nonce.SenderValue, ScKeys[peerNodeId].Personalization);
            int sliceindex = nonce.GetSliceIndex();
            var lastSliceIndex = spanCache.CtxSpanCacheSlices.Count - 1;
            CTR_DRBG_CTX ctx;
            if (sliceindex > lastSliceIndex)
            {
                if (spanCache.CtxSpanCacheSlices.Count == 0)
                {
                    ctx = SpanTable.GetContainer(peerNodeId).Context.Clone();
                    spanCache.CtxSpanCacheSlices.Add(ctx);
                    lastSliceIndex = 0;
                }
                else
                {
                    ctx = spanCache.CtxSpanCacheSlices[lastSliceIndex];
                    SpanTable.GetContainer(peerNodeId).Context = ctx.Clone();
                }
                for (int i = lastSliceIndex; i < sliceindex; i++)
                {
                    for (int j = 0; j < nonce.GetSliceSpan(); j++)
                    {
                        spanContainer.NextNonce();
                    }
                    ctx = SpanTable.GetContainer(peerNodeId).Context.Clone();
                    spanCache.CtxSpanCacheSlices.Add(ctx);
                }
            }
            else
            {
                ctx = spanCache.CtxSpanCacheSlices[sliceindex];
                SpanTable.GetContainer(peerNodeId).Context = ctx.Clone();
            }
            var leftIterations = nonce.GetSliceLeftIterations();
            for (int j = 0; j < leftIterations - 1; j++)
            {
                spanContainer.NextNonce();
            }
            bool isDecrypted = false;
            for (int i = 0; i < 1 + (nonce.Generation / 10); i += 5)
            {
                isDecrypted = SecurityS2CryptoProviderBase.DecryptSinglecastFrame(
                    spanContainer,
                    ScKeys.ContainsKey(peerNodeId) ? ScKeys[peerNodeId] : null,
                    dataItem.Source,
                    dataItem.Destination[0],
                    dataItem.HomeId,
                    dataItem.LastPayload,
                    out payload,
                    out extensions);

                if (isDecrypted)
                {
                    break;
                }
            }
            return isDecrypted;
        }

        public static IEnumerable<byte[]> GetTempSecurityKeys()
        {
            return _securityKeys;
        }

        public static IEnumerable<byte[]> GetSecurityKeys()
        {
            return _tempSecurityKeys;
        }

        public static void AddSecurityKey(byte[] keyValue)
        {
            if (!_securityKeys.Where(x => x.SequenceEqual(keyValue)).Any())
            {
                _securityKeys.Add(keyValue);
            }
        }

        public static void AddSecurityKeys(List<byte[]> networkKeys)
        {
            if (networkKeys != null)
            {
                foreach (var keyValue in networkKeys)
                {
                    AddSecurityKey(keyValue);
                }
            }
        }

        public static void AddTempSecurityKey(byte[] keyValue)
        {
            if (!_tempSecurityKeys.Where(x => x.SequenceEqual(keyValue)).Any())
            {
                _tempSecurityKeys.Add(keyValue);
            }
        }

        public static void AddTempSecurityKeys(List<byte[]> tempNetworkKeys)
        {
            if (tempNetworkKeys != null)
            {
                foreach (var keyValue in tempNetworkKeys)
                {
                    AddTempSecurityKey(keyValue);
                }
            }
        }
    }

    public class NonceS0Cache
    {
        public NonceS0Cache()
        {
            NetworkKeyIndex = -1;
        }

        public int NetworkKeyIndex { get; set; }
        public int ProbedNetworkKeyCount { get; set; }
    }

    public class SpanCache
    {
        public SpanCache()
        {
            NetworkKeyIndex = -1;
            TempNetworkKeyIndex = -1;
            CtxSpanCacheSlices = new List<CTR_DRBG_CTX>();
        }

        public int NetworkKeyIndex { get; set; }
        public int ProbedNetworkKeyCount { get; set; }
        public int TempNetworkKeyIndex { get; set; }
        public int ProbedTempNetworkKeyCount { get; set; }

        /// <summary>
        /// Stored slices every 1K iteration
        /// </summary>
        public List<CTR_DRBG_CTX> CtxSpanCacheSlices { get; set; }
    }

    public class MpanCache
    {
        public MpanCache()
        {
            NetworkKeyIndex = -1;
            CtxMpanCacheSlices = new List<CTR_DRBG_CTX>();
        }

        public int NetworkKeyIndex { get; set; }
        public int ProbedNetworkKeyCount { get; set; }

        /// <summary>
        /// Stored slices every 1K iteration
        /// </summary>
        public List<CTR_DRBG_CTX> CtxMpanCacheSlices { get; set; }
    }

    public class SpanRecord
    {
        public int SyncId { get; set; }
        public int Generation { get; set; }
        public byte[] ReceiverValue { get; set; }
        public byte[] SenderValue { get; set; }
        private byte _sequenceNumberLeft; // seqNo from node with lesser source nodeId (not persistent) 
        private byte _sequenceNumberRight; // seqNo from node with grater source nodeId (not persistent)

        public static implicit operator SpanRecord(byte[] data)
        {
            SpanRecord ret = new SpanRecord();
            if (data != null)
            {
                if (data.Length == 24)
                {
                    ret.SyncId = BitConverter.ToInt32(data, 0);
                    ret.Generation = BitConverter.ToInt32(data, 4);
                    ret.ReceiverValue = data.Skip(8).Take(16).ToArray();
                }
                else if (data.Length == 40)
                {
                    ret.SyncId = BitConverter.ToInt32(data, 0);
                    ret.Generation = BitConverter.ToInt32(data, 4);
                    ret.ReceiverValue = data.Skip(8).Take(16).ToArray();
                    ret.SenderValue = data.Skip(24).Take(16).ToArray();
                }
            }
            return ret;
        }

        public static implicit operator byte[] (SpanRecord value)
        {
            List<byte> ret = new List<byte>();
            ret.AddRange(BitConverter.GetBytes(value.SyncId));
            ret.AddRange(BitConverter.GetBytes(value.Generation));
            ret.AddRange(value.ReceiverValue);
            ret.AddRange(value.SenderValue);
            return ret.ToArray();
        }

        public bool IsValidSequenceNumber(byte src, byte dest, byte sequenceNumber)
        {
            bool ret = false;
            if (Generation == 0)
            {
                ret = true;
            }
            else
            {
                if (src > dest)
                {
                    ret = _sequenceNumberRight != sequenceNumber;
                }
                else
                {
                    ret = _sequenceNumberLeft != sequenceNumber;
                }
            }
            return ret;
        }

        public void SetSequenceNumber(byte src, byte dest, byte sequenceNumber)
        {
            if (src > dest)
            {
                _sequenceNumberRight = sequenceNumber;
            }
            else
            {
                _sequenceNumberLeft = sequenceNumber;
            }
        }

        public int GetSliceSpan()
        {
            return 1000;
        }

        public int GetSliceIndex()
        {
            return Generation / GetSliceSpan();
        }

        public int GetSliceLeftIterations()
        {
            return Generation % GetSliceSpan();
        }
    }

    public class MpanValue
    {
        public int Position { get; set; }
        public byte[] Value { get; set; }
        public int KeyIndex { get; set; }
    }
}
