﻿using System;
using System.Linq;
using NUnit.Framework;
using ZWaveController.ViewModels;
using ZWave.BasicApplication.Operations;
using ZWave.BasicApplication.Devices;
using ZWave.Enums;
using ZWave.Devices;
using ZWave.CommandClasses;
using Utils.UI;
using ZWave.BasicApplication.EmulatedLink;
using ZWave.BasicApplication;
using ZWave.Layers.Session;
using ZWave.Layers;

namespace Tests
{
    /// <summary>
    /// The Command Classes Tests class.
    /// Test Send Data for Command Classes View Model in PCController.
    /// </summary>
    [TestFixture]
    public class CommandClassesTests
    {
        private const byte PRIMARY_NODE_ID = 0x01;
        private const byte SECONDRY_NODE_ID = 0x02;
        private byte[] HOME_ID = new byte[] { 0xDE, 0xAD, 0xBE, 0xEF };

        private MainViewModel _mainVM;
        private Controller _controllerPrimary;
        private Controller _controllerSecondary;

        protected BasicLinkTransportLayer _transport;
        protected BasicApplicationLayer _app;

        private void InitMainViewModel(Controller controller1)
        {
            var dispatcherFake = new DirectDispatcher();
            var subscribeCollectionFactoryStub = new UListFactory();
            _mainVM = new MainViewModel(subscribeCollectionFactoryStub, dispatcherFake);
            //Load command classes definition to write to log
            _mainVM.Controller = controller1;
            _mainVM.LoadZWaveDefinition(@".\XmlFiles\ZWave_custom_cmd_classes.xml");
        }

        private void SetSelectedNode()
        {
            var di = new NodeTag(SECONDRY_NODE_ID);
            _mainVM.Invoke(() => _mainVM.ConfigurationItem.AddOrUpdateNode(di));
            _mainVM.SelectedNode = _mainVM.ConfigurationItem.Nodes.Last();
        }

        private string Data2HexString(byte[] data)
        {
            return string.Concat(string.Join(" ", (data.Select(x => x.ToString("X2")).ToArray())));
        }

        [SetUp]
        public void SetUp()
        {
            _transport = new BasicLinkTransportLayer();
            _app = new BasicApplicationLayer(new SessionLayer(), new BasicFrameLayer(), _transport);

            _controllerPrimary = _app.CreateController();
            _controllerSecondary = _app.CreateController();

            _controllerPrimary.Connect(new SerialPortDataSource("COM1"));
            _controllerSecondary.Connect(new SerialPortDataSource("COM2"));

            _transport.SetUpModulesNetwork(_controllerPrimary.SessionId, _controllerSecondary.SessionId);

            _controllerPrimary.MemoryGetId();
            _controllerSecondary.MemoryGetId();

            InitMainViewModel(_controllerPrimary);
            SetSelectedNode();
        }

        [TearDown]
        public void TearDown()
        {
            _controllerPrimary.Dispose();
            _controllerSecondary.Dispose();
        }

        [Test]
        public void SendData_FromPrimaryToSecondary_SecondaryReceivesData()
        {
            int i = 1;
            foreach (var commandClass in _mainVM.ZWaveDefinition.CommandClasses)
            {
                Console.WriteLine("-Command Classs Name: {0}", commandClass.Name);
                if (commandClass.Command == null ||
                    commandClass.Name.Contains("COMMAND_CLASS_ZIP")
                    )
                {
                    continue;
                }
                foreach (var command in commandClass.Command)
                {
                    Console.WriteLine("[+] Iteration N{0}, Command: {1}", i++, command.Name);
                    _mainVM.CommandClassesViewModel.SelectedCommand = command;
                    Console.WriteLine(" | Before updating parameters: {0}", Data2HexString(_mainVM.CommandClassesViewModel.Payload));
                    if (_mainVM.CommandClassesViewModel.Parameters != null && _mainVM.CommandClassesViewModel.Parameters.Count > 0)
                    {
                        _mainVM.CommandClassesViewModel.Parameters[0].Value = _mainVM.CommandClassesViewModel.Parameters[0].Value;
                    }
                    Console.WriteLine(" | After updating parameters: {0}", Data2HexString(_mainVM.CommandClassesViewModel.Payload));
                    var data = _mainVM.CommandClassesViewModel.Payload;

                    var expectToken = _controllerSecondary.ExpectData(data, 1000, null);
                    _controllerPrimary.SendData(SECONDRY_NODE_ID, data, TransmitOptions.TransmitOptionAcknowledge);
                    var expectRes = (ExpectDataResult)expectToken.WaitCompletedSignal();
                    Assert.IsNotNull(data);
                    Assert.IsNotNull(expectRes);
                    Assert.IsNotNull(expectRes.Command);
                    Assert.IsTrue(expectRes.Command.SequenceEqual(data), command.Name);
                }
            }
        }
    }
}
